// fmul.cpp: parallel floating-point benchmark
//
// modifications:
//
// 09/30/16 RD	use 100 million iterations, 32 units (release v030)
// 05/31/16 RD	initial SystemC benchmark (derived from SpecC example)

#include <stdio.h>
#include <stdlib.h>
#include "systemc.h"

// number of parallel instances (powers of 2 in range 1 to 1024)
#ifndef PAR
#define PAR 32
#endif

// number of multiplications per unit
#define MAXLOOP 100000000

// type of floating-point numbers
typedef double fp_t;

SC_MODULE(Fmul1)
{
	int i;
	fp_t f;

	void main(void)
	{
		i = 0;
		f = 1.2;
		while(i < MAXLOOP)
		{
			f *= 1.1;
			i ++;
		}
	}

	SC_CTOR(Fmul1)
	{
		SC_THREAD(main);
	}
};

#define FMUL_PAR(NAME, COMPONENT)		\
SC_MODULE(NAME)					\
{						\
	COMPONENT	fmul1,			\
			fmul2;			\
						\
	SC_CTOR(NAME):				\
		fmul1("fmul1"),			\
		fmul2("fmul2")			\
	{					\
	}					\
};

#if PAR >= 2
FMUL_PAR(Fmul2, Fmul1)
#define TOP Fmul2
#endif

#if PAR >= 4
FMUL_PAR(Fmul4, Fmul2)
#undef TOP
#define TOP Fmul4
#endif

#if PAR >= 8
FMUL_PAR(Fmul8, Fmul4)
#undef TOP
#define TOP Fmul8
#endif

#if PAR >= 16
FMUL_PAR(Fmul16, Fmul8)
#undef TOP
#define TOP Fmul16
#endif

#if PAR >= 32
FMUL_PAR(Fmul32, Fmul16)
#undef TOP
#define TOP Fmul32
#endif

#if PAR >= 64
FMUL_PAR(Fmul64, Fmul32)
#undef TOP
#define TOP Fmul64
#endif

#if PAR >= 128
FMUL_PAR(Fmul128, Fmul64)
#undef TOP
#define TOP Fmul128
#endif

#if PAR >= 256
FMUL_PAR(Fmul256, Fmul128)
#undef TOP
#define TOP Fmul256
#endif

#if PAR >= 512
FMUL_PAR(Fmul512, Fmul256)
#undef TOP
#define TOP Fmul512
#endif

#if PAR >= 1024
FMUL_PAR(Fmul1024, Fmul512)
#undef TOP
#define TOP Fmul1024
#endif

TOP	top("top");

int sc_main(int argc, char* argv[])
{
	printf("Fmul[%d,%d] starting...\n", PAR, MAXLOOP);
	sc_start();
	printf("Done!\n");
	return(0);
}

// EOF fmul.cpp
