.TH risc l "July 31, 2017"
.SH NAME
.B risc
\- Recoding Infrastructure for SystemC (RISC) Compiler and Simulator
.SH SYNOPSIS
.B risc
[
.I options
]
.I design
[
.I options
]
.SH DESCRIPTION
.B risc
is a dedicated compiler for the SystemC language.
The purpose of
.B risc
is to parse, analyze, instrument, and compile a SystemC source program
into an executable program for out-of-order parallel simulation.
.B risc
is a frontend source-to-source compiler for SystemC built on top of
the ROSE compiler infrastructure with GNU or Intel C++
as backend target compiler.
As such,
.B risc
relies on and supports also most of the ROSE and GNU compiler options.

Using the command syntax shown in the synopsis above, the specified
.I design
is compiled. By default,
.B risc
reads the SystemC source file, performs preprocessing and
builds an internal representation (abstract syntax tree)
and a Segment Graph (SG) of the model. Next, segment conflict analysis
is performed and the design model is instrumented
for Out-of-Order Parallel Discrete Event Simulation (OoO PDES). Finally,
instrumented C++ code is generated, compiled, and linked
into an executable file that can be run for fast parallel simulation.

On successful completion, the exit value 0 is returned.
In case of errors during processing, an error code with a
brief diagnostic message is written to
the standard error stream and the compilation is aborted
with an exit value greater than zero.

For preprocessing and C++ compilation into an executable file,
.B risc
relies on the availability of an external C++ compiler which
is used automatically in the background.
By default, the GNU C++ compiler
.B g++
is used. Alternatively (see options
.I \-risc:icpc
and 
.I \-risc:mic
below), the Intel C++ compiler
.B icpc
may be used to generate an executable optimized for Intel processors
with SIMD capabilities or the Intel Many-Integrated-Core (MIC) architecture.
.SH ARGUMENTS
.TP
.I design
specifies the file name of the input SystemC design model;
by default, the base name of
.I design
is used as base name for the intermediate and output files;
.SH OPTIONS
.TP
.I \-h | \-\-help
print the
.B risc
compiler version and a brief usage information message
to standard output and quit;
.TP
.I \-v | \-\-verbose
increment the verbosity level so that all tasks performed are
logged to standard error (default: be silent);
at level 1, high-level messages about the tasks performed are displayed;
at level 2, additional details such as input and output file names are listed;
at level 3, very detailed information about each executed task is printed;
.TP
.I \-vv
increment the verbosity level by two counts (same as
.I \-v \-v
);
.TP
.I \-vvv
increment the verbosity level by three counts (same as
.I \-v \-v \-v
);
.TP
.I \-w | \-\-warnings
increment the warning level so that compiler warning messages
are enabled (default: warnings are disabled);
four levels are supported ranging
from only important warnings (level 1)
to pedantic warnings (level 4);
for most cases, warning level 2 is recommended (
.I \-w \-w
);
.TP
.I \-ww
increment the warning level by two counts (same as
.I \-w \-w
);
.TP
.I \-www
increment the warning level by three counts (same as
.I \-w \-w \-w
);
.TP
.I \-g
add a symbol table suitable for debugging (e.g. using
.B gdb
) to the generated object files and
simulation executable (default: no debugging symbols);
.TP
.I \-O | \-O level
optimize the generated simulation executable
for higher execution speed and/or less memory usage (default: no optimization);
.TP
.I \-Idir
add the specified
.I dir
to the include path (extend the list of directories to be searched
for including source files);
include directories are searched in the order of their specification;
the standard include path ($SYSTEMC_LW_HOME/include
or $SYSTEMC_OOP_HOME/include) is
automatically appended to this list;
by default, only the standard include directories are searched;
.TP
.I \-Ldir
add the specified
.I dir
to the library path (extend the list of directories to be searched
for linker libraries);
the library path is searched in the specified order;
the standard library path ($SYSTEMC_OOP_HOME/lib) is
automatically appended to this list;
by default, only the standard library path is searched;
.TP
.I \-llib
add the specified
.I lib
to the list of libraries for the linker
so that the executable is linked against
.I lib;
libraries are linked in the specified order;
the standard libraries (i.e. -lsystemc) are
automatically appended to this list;
by default, only standard libraries are used;
.TP
.I \-c
perform only the preprocessing, analysis, instrumentation,
and compilation tasks; skip the final linking stage
so that only an object file is created (default: perform all tasks
including linking);
.TP
.I \-o output file
specify the name of the final output file explicitly (default: a.out);
.TP
.I \-risc:dump
output the computed segment graph (SG) and conflict tables as HTML
files (default: no HTML files are generated); these files may be viewed
by a user in a browser in order to inspect the out-of-order execution
conditions of the model and improve it accordingly;
.TP
.I \-risc:icpc
use the Intel C++ compiler
.B icpc
in the backend for generating the executable (default: GNU C++ compiler
.B g++
);
.TP
.I \-risc:mic
use the Intel C++ compiler
.B icpc
with option
.I \-mic
in the backend for cross-compiling an executable for the Intel
Many Integrated Core (MIC) architecture (default: generate an executable
for the same processor the compiler is running on);
.TP
.I \-risc:elab filename
import the elaboration result produced by the RISC elaborator
.B elab
from file
.I filename
and use it for segment conflict analysis based on a dynamic elaboration phase
(default: pure static analysis);
.TP
.I \-<rose:option>
pass this option through to the underlying ROSE compiler (default: none);
.TP
.I \-<GNU option>
pass this option through to the underlying GNU compiler (default: none);
.SH ENVIRONMENT
.TP
.I RISC
is used at compile-time to determine the installation directory
of the RISC compiler and simulator where the RISC system components
are located (default: none);
.TP
.I SYSTEMC_LW_HOME
is used at compile-time to find the RISC light-weight SystemC header files
which are expected in directory $SYSTEMC_LW_HOME/include (default: none);
.TP
.I SYSTEMC_OOP_HOME
is used at compile-time to find the RISC out-of-order SystemC header files
which are expected in directory $SYSTEMC_OOP_HOME/include,
and the RISC out-of-order SystemC library which is expected in
directory $SYSTEMC_OOP_HOME/lib (default: none);
.TP
.I SYSTEMC_MIC_HOME
is used at compile-time to find the RISC SystemC header files
and library files for the Intel many-integrated-core (MIC) architecture
which are expected in directory $SYSTEMC_MIC_HOME/include and
and $SYSTEMC_MIC_HOME/lib, respectively (default: none);
this is used only when the option
.I \-mic
is used (see above);
.TP
.I SYSC_PRINT_MODE_MESSAGE
is used by the RISC simulator at run-time to print the mode of simulation
and also the actual values of the environment variables listed below;
these log lines start with "***" and are only printed
when $SYSC_PRINT_MODE_MESSAGE is defined (default: no messages are printed);
.TP
.I SYSC_SYNC_PAR_SIM
is used by the RISC simulator at run-time to force the RISC out-of-order
SystemC simulation to fall back to synchronous (in-order) PDES execution;
note that this mode is also automatically selected when SystemC primitive
channels with update requests are used (default: out-of-order execution);
.TP
.I SYSC_PRINT_VERBOSE_MESSAGE
is used by the RISC simulator at run-time to print debugging information
about the simulator queues, event processing, and time advances;
such debugging lines are only printed
when $SYSC_PRINT_VERBOSE_MESSAGE is defined (default: no debugging infos
are printed);
.TP
.I SYSC_DISABLE_PREDICTION
is used by the RISC simulator at run-time to switch back
to non-predictive conflict detection; this avoids scheduling overhead
at run time, but usually results in slower simulation due to more conflicts;
if $SYSC_DISABLE_PREDICTION is defined, thread state prediction is not used
during out-of-order scheduling (default: out-of-order execution
with prediction);
.TP
.I SYSC_PAR_SIM_CPUS
is used by the RISC simulator at run-time to set the maximum number
of concurrent threads allowed in the RISC out-of-order SystemC
simulation (default: 64);
.SH VERSION
The RISC compiler and simulator is release version 0.4.0.
.SH AUTHORS
Zhongqi Cheng <zhongqc@uci.edu>,
Rainer Doemer <doemer@uci.edu>,
Guantao Liu <guantaol@uci.edu>, and
Tim Schmidt <schmidtt@uci.edu>.
.SH COPYRIGHT
(c) 2017 CECS, University of California, Irvine
.SH LICENSE
Open source BSD license terms apply.
.SH BUGS, LIMITATIONS
This is an academic proof-of-concept prototype implementation,
not commercial-quality software.
See the file BUGS in the software packages for known limitations.

