#include <iostream>

#include "risc.h"
#include <list>
#include <ctime>

#include "earliest_execution_table.h"

struct Lookup {

  Lookup(risc::InstanceTree path, risc::Function *function):
    path_(path),
    function_(function)
  { }

  risc::InstanceTree path_;
  risc::Function *function_;
};

risc::InstanceTree function_call_to_instance_tree_lookup();

SgFunctionDeclaration*
get_virtual_function_declaration(SgFunctionCallExp *func_call);

void
get_all_instances_of(
    std::list<risc::InstanceTree> &instances,
    risc::InstanceTree root,
    risc::Definition::Type type,
    bool first);

std::list<risc::sg::VertexDescriptor>
get_receiving_segments(risc::InstanceTree channel, risc::sg::SegmentGraph &sg,
    std::list<Lookup> function_instance_lookup);

std::list<risc::sg::VertexDescriptor>
get_receiving_segments(risc::InstanceTree channel, risc::sg::SegmentGraph &sg,
    std::list<Lookup> function_instance_lookup)
{
//  std::cout << "Channel: ";
  channel.print_instance_tree();
  std::list<risc::sg::VertexDescriptor> result;

  std::pair<risc::sg::VertexIterator, risc::sg::VertexIterator> vp;
  for(vp = boost::vertices(sg.graph_);
      vp.first != vp.second; ++vp.first) {

    for(risc::sg::Segment::Expressions::iterator
        iter_segment  = sg.graph_[*(vp.first)].expressions_.begin();
        iter_segment != sg.graph_[*(vp.first)].expressions_.end();
        iter_segment++) {

      SgFunctionCallExp *func_call = isSgFunctionCallExp(*iter_segment);

      if(func_call) {

        SgFunctionDeclaration *func_decl =
          func_call->getAssociatedFunctionDeclaration();

        if(!func_decl) {

          func_decl = get_virtual_function_declaration(func_call);

        } else {

          if(SageInterface::get_name(func_decl) != "read") {
            continue;
          }
        }

        if(func_decl) {

          if(SageInterface::get_name(func_decl) != "read") {
            continue;
          }

          // getting the associated port
          SgFunctionDefinition *func_def
            = SageInterface::getEnclosingFunctionDefinition(func_call);

          risc::InstanceTree module;

          for(std::list<Lookup>::iterator
              iter_lookup  = function_instance_lookup.begin();
              iter_lookup != function_instance_lookup.end();
              iter_lookup++) {

            std::list<int> all_reachable_segments
              = get_all_reachable_segments(sg,
                  (*iter_lookup).function_->starting_id_);

            std::list<int>::iterator path_iter
              = std::find(all_reachable_segments.begin(),
                  all_reachable_segments.end(),
                  sg.graph_[(*vp.first)].id_);

            if(path_iter != all_reachable_segments.end()) {

              // getting the port in the function call
              SgArrowExp *arrow_exp = isSgArrowExp(func_call->get_function());
              SgFunctionCallExp *call_exp
                = isSgFunctionCallExp(arrow_exp->get_lhs_operand());
              SgDotExp *dot_exp = isSgDotExp(call_exp->get_function());
              SgCastExp *cast_exp = isSgCastExp(dot_exp->get_lhs_operand());
              SgArrowExp *arrow_exp2 = isSgArrowExp(cast_exp->get_operand());
              SgVarRefExp *var_ref
                = isSgVarRefExp(arrow_exp2->get_rhs_operand());
              SgVariableDefinition *var_def
                = isSgVariableDefinition(
                    var_ref->get_symbol()->get_declaration()->get_definition());

              // getting the path to the port
              risc::InstanceTree tree_to_function = (*iter_lookup).path_;
              risc::Object *obj = tree_to_function.get_object();

              risc::ModuleInstance *mod = dynamic_cast<risc::ModuleInstance*>(obj);
              risc::Port *port = mod->get_definition()->find_port(var_def);
              risc::InstanceTree path_to_port = (*iter_lookup).path_;

              // Adding the port element to the path
              risc::InstanceTree::Element port_element;
              port_element.first = mod->get_definition();
              port_element.second = port;
              path_to_port.path_.push_back(port_element);


              risc::InstanceTree mapped_channel
                = risc::get_mapped_channel_instance(path_to_port);

              if(channel == mapped_channel) {

                result.push_back(*vp.first);
              }
              // looking for all receivers
            }
          }
        }
      }
    }
  }
  return result;
}

SgFunctionDeclaration*
get_virtual_function_declaration(SgFunctionCallExp *func_call)
{
  if(!isSgArrowExp(func_call->get_function())) {
    return NULL;
  }

  SgMemberFunctionRefExp *member_func
    = isSgMemberFunctionRefExp(
          isSgArrowExp(func_call->get_function())->get_rhs_operand());

  std::vector<SgNode*> class_defs
    = NodeQuery::querySubTree(risc::get_project(), V_SgClassDefinition);

  std::vector<SgNode*> func_decls;
  for(std::vector<SgNode*>::iterator
      iter  = class_defs.begin();
      iter != class_defs.end();
      iter++) {

    SgClassDefinition *class_def = isSgClassDefinition(*iter);

    if(risc::tools::is_derived_from_sc_interface(class_def)) {

      std::vector<SgNode*> func_decls_temp
        = NodeQuery::querySubTree(class_def, V_SgFunctionDeclaration);

      func_decls.insert(func_decls.end(), func_decls_temp.begin(),
          func_decls_temp.end());
    }
  }

  for(std::vector<SgNode*>::iterator
      iter  = func_decls.begin();
      iter != func_decls.end();
      iter++) {

    SgFunctionDeclaration *func_decl = isSgFunctionDeclaration(*iter);

    func_decl = isSgFunctionDeclaration(func_decl->get_definingDeclaration());

    if(!func_decl) {
      continue;
    }

    if(member_func->unparseToString() != SageInterface::get_name(func_decl)) {

      continue;
    }

    if(func_decl->get_orig_return_type() != func_call->get_type()) {

      continue;
    }

    if(func_decl->get_parameterList()->get_args().size()
        != func_call->get_args()->get_expressions().size()) {
      continue;
    }

    bool equal = true;
    for(int i = 0; i < func_decl->get_parameterList()->get_args().size(); i++) {


      if(func_decl->get_args().at(i)->get_type()
          != func_call->get_args()->get_expressions().at(i)->get_type()) {

        if(isSgReferenceType(func_decl->get_args().at(i)->get_type())
            && !isSgReferenceType(
                    func_call->get_args()->get_expressions().at(i)->get_type())
            && !isSgPointerType(
                    func_call->get_args()->get_expressions().at(i)->get_type()) 
            && (isSgReferenceType(
                    func_decl->get_args().at(i)->get_type())->get_base_type()
              != func_call->get_args()->get_expressions().at(i)->get_type())) {
          equal = false;
        }
      }
    }

    if(!equal) {
      continue;
    }

    return func_decl;
  }
  return NULL;
}


int main(int argc, char* argv[])
{
  std::cout << std::endl
            << "start processing file: " __FILE__
            << " ..."
            << std::endl;

  risc::Design d(argc,argv);

  d.print_design();


  risc::InstanceTree tree = d.get_root();

  std::list<risc::InstanceTree> ports;
  tree.get_all_ports(ports);
  std::cout << "All ports:" << std::endl;
  for(std::list<risc::InstanceTree>::iterator
      iter  = ports.begin();
      iter != ports.end();
      iter++) {

    (*iter).print_instance_tree();

    risc::InstanceTree mapped_channel = get_mapped_channel_instance(*iter);
    std::cout << " -> " << std::endl;
    mapped_channel.print_instance_tree();
    std::cout << std::endl;
    std::cout << std::endl;
  }



  std::cout << "=========================================" << std::endl;

  // Creating a segment graph
  risc::sg::SegmentGraph::FunctionCallBoundarySet
    function_call_boundary(
        risc::tools::SystemCDefinitions::get_sc_wait_funcs());

  risc::sg::SegmentGraph sg(function_call_boundary);

  std::list<risc::InstanceTree> channel_instances;
  get_all_instances_of(channel_instances,
      d.get_root(),
      risc::Definition::PrimitiveChannelInstanceType,
      true);


  for(std::list<risc::InstanceTree>::iterator
      iter  = channel_instances.begin();
      iter != channel_instances.end();
      iter++) {

    //risc::PrimitiveChannelInstance *channel_instance
    //  = dynamic_cast<risc::ModuleInstance*>((*iter).get_object());

    (*iter).print_instance_tree();
  }

  //////////////////////////////////////////////////////////
  std::list<risc::InstanceTree> module_instances;
  get_all_instances_of(module_instances,
      d.get_root(),
      risc::Definition::ModuleInstanceType,
      true);

  //////////////////////////////////////////////////////////

  std::list<Lookup> function_instance_lookup;

  for(std::list<risc::InstanceTree>::iterator
      iter  = module_instances.begin();
      iter != module_instances.end();
      iter++) {

    risc::ModuleInstance *module_instance
      = dynamic_cast<risc::ModuleInstance*>((*iter).get_object());


    std::cout << "Module: " << module_instance->get_name()
              << ", "
              << module_instance->get_definition()->get_name()
              << std::endl;

    std::cout << "Threads: "
              << module_instance->get_definition()->threads_.size()
              << std::endl;

    std::cout << "CThreads: "
              << module_instance->get_definition()->cthreads_.size()
              << std::endl;

    std::cout << "Methods: "
              << module_instance->get_definition()->methods_.size()
              << std::endl;

    std::cout << std::endl;

    for(risc::ThreadVectorIter
        iter_thread  = module_instance->get_definition()->threads_.begin();
        iter_thread != module_instance->get_definition()->threads_.end();
        iter_thread++) {

      if((*iter_thread)->get_name() == "thread1" ) {

        AstDOTGeneration dotgen;
        dotgen.generate((*iter_thread)->get_ast_node(), (*iter_thread)->get_name());
      }

      sg.add_function(*iter_thread, true);

      std::cout << "Adding thread: " << (*iter_thread)->get_name() << std::endl;
      function_instance_lookup.push_back(Lookup(*iter, *iter_thread));
    }


    for(risc::CThreadVectorIter
        iter_thread  = module_instance->get_definition()->cthreads_.begin();
        iter_thread != module_instance->get_definition()->cthreads_.end();
        iter_thread++) {

      std::cout << "Adding cthread: " << (*iter_thread)->get_name() << std::endl;
      sg.add_function(*iter_thread, true);

      //std::cout << "Adding cthread: " << (*iter_thread)->get_name() << std::endl;

      function_instance_lookup.push_back(Lookup(*iter, *iter_thread));
    }
  }

  std::cout << "lalala" << std::endl;

  sg.clean_graph();
  sg.print_graph("sg.dot");

  //////////////////////////////////////////////////////////
  // Adding event notification edges
  risc::sg::EventConflictTable
    event_conflict_table(sg, d.path_instance_mapper_, true);

  std::list<risc::sg::EdgeDescriptor> event_edges;

  for(int
      row = 0;
      row < event_conflict_table.size_of_conflict_table_;
      row++) {

    for(int
        column = 0;
        column < event_conflict_table.size_of_conflict_table_;
        column++) {

      if(event_conflict_table.conflict_table_[row][column]) {

        std::pair<int, int> seg_and_inst_id_column
          = event_conflict_table.index_to_segment_and_instance_id(column);

        std::pair<int, int> seg_and_inst_id_row
          = event_conflict_table.index_to_segment_and_instance_id(row);

        std::pair<risc::sg::EdgeDescriptor, bool> edge =
          boost::add_edge(
              event_conflict_table.
                segment_id_to_vertex_descriptor(seg_and_inst_id_row.first),
              event_conflict_table.
                segment_id_to_vertex_descriptor(seg_and_inst_id_column.first),
              sg.graph_);

        if(edge.second) {
          sg.graph_[edge.first].event_notification_edge_ = true;
          event_edges.push_back(edge.first);
        }
      }
    }
  }

  sg.print_graph("event_graph.dot");

  for(std::list<risc::sg::EdgeDescriptor>::iterator
      iter  = event_edges.begin();
      iter != event_edges.end();
      iter++) {
//    boost::remove_edge(*iter, sg.graph_);
  }

  //////////////////////////////////////////////////////////
  // Adding communication notification edges

  std::cout << "Begin of communication edges" << std::endl;

  std::pair<risc::sg::VertexIterator, risc::sg::VertexIterator> vp;
  for(vp = boost::vertices(sg.graph_);
      vp.first != vp.second; ++vp.first) {

    for(risc::sg::Segment::Expressions::iterator
        iter_segment  = sg.graph_[*(vp.first)].expressions_.begin();
        iter_segment != sg.graph_[*(vp.first)].expressions_.end();
        iter_segment++) {

      SgFunctionCallExp *func_call = isSgFunctionCallExp(*iter_segment);

      if(func_call) {

        SgFunctionDeclaration *func_decl =
          func_call->getAssociatedFunctionDeclaration();

        if(!func_decl) {

          func_decl = get_virtual_function_declaration(func_call);

        } else {

          if(SageInterface::get_name(func_decl) != "write") {
            continue;
          }
        }

        if(func_decl) {

          if(SageInterface::get_name(func_decl) != "write") {
            continue;
          }

          // getting the associated port
          SgFunctionDefinition *func_def
            = SageInterface::getEnclosingFunctionDefinition(func_call);

          risc::InstanceTree module;

          for(std::list<Lookup>::iterator
              iter_lookup  = function_instance_lookup.begin();
              iter_lookup != function_instance_lookup.end();
              iter_lookup++) {

            std::list<int> all_reachable_segments
              = get_all_reachable_segments(sg,
                  (*iter_lookup).function_->starting_id_);

            std::list<int>::iterator path_iter
              = std::find(all_reachable_segments.begin(),
                  all_reachable_segments.end(),
                  sg.graph_[(*vp.first)].id_);

            if(path_iter != all_reachable_segments.end()) {

              // getting the port in the function call
              SgArrowExp *arrow_exp = isSgArrowExp(func_call->get_function());

              if(!arrow_exp)
                continue;

              SgFunctionCallExp *call_exp
                = isSgFunctionCallExp(arrow_exp->get_lhs_operand());

              if(!call_exp)
                continue;

              SgDotExp *dot_exp = isSgDotExp(call_exp->get_function());
              SgCastExp *cast_exp = isSgCastExp(dot_exp->get_lhs_operand());
              SgArrowExp *arrow_exp2 = isSgArrowExp(cast_exp->get_operand());
              SgVarRefExp *var_ref
                = isSgVarRefExp(arrow_exp2->get_rhs_operand());
              SgVariableDefinition *var_def
                = isSgVariableDefinition(
                    var_ref->get_symbol()->get_declaration()->get_definition());

              // getting the path to the port
              risc::InstanceTree tree_to_function = (*iter_lookup).path_;
              risc::Object *obj = tree_to_function.get_object();

              risc::ModuleInstance *mod
                = dynamic_cast<risc::ModuleInstance*>(obj);

              risc::Port *port = mod->get_definition()->find_port(var_def);
              risc::InstanceTree path_to_port = (*iter_lookup).path_;

              // Adding the port element to the path
              risc::InstanceTree::Element port_element;
              port_element.first = mod->get_definition();
              port_element.second = port;
              path_to_port.path_.push_back(port_element);

              risc::InstanceTree mapped_channel
                = risc::get_mapped_channel_instance(path_to_port);

              std::list<risc::sg::VertexDescriptor> reading_segments
                = get_receiving_segments(mapped_channel, sg,
                    function_instance_lookup);

              reading_segments.sort();
              reading_segments.unique();

              int i =0;
              for(std::list<risc::sg::VertexDescriptor>::iterator
                  iter  = reading_segments.begin();
                  iter != reading_segments.end();
                  iter++) {

                std::pair<risc::sg::EdgeDescriptor, bool> edge =
                  boost::add_edge(
                      *(vp.first), *iter,
                      sg.graph_);

                if(edge.second) {
                  sg.graph_[edge.first].hierarchical_communication_ = true;
                }
              }

              // looking for all receivers
            }
          }
        }
      }
    }
  }







  sg.print_graph("communication_graph.dot");

  std::cout << "++++++++++++++++++++++++++++" << std::endl;

  std::cout << std::endl;
  std::cout << "Print table" << std::endl;

  std::cout << std::endl;
  event_conflict_table.print_conflict_table("event_conflicts.html");

  std::cout << "end processing file: " __FILE__
            << " ..."
            << std::endl
            << std::endl;

  return 0;
}

// Assumption: only channels in channels and modules in modules
void get_all_instances_of(std::list<risc::InstanceTree> &instances,
         risc::InstanceTree root, risc::Definition::Type type, bool first)
{
  risc::ModuleInstance *module_instance
    = dynamic_cast<risc::ModuleInstance*>(root.get_object());

  if(module_instance) {

    for(risc::ModuleInstanceVectorIter
        iter  = module_instance->get_definition()->sub_modules_.begin();
        iter != module_instance->get_definition()->sub_modules_.end();
        iter++) {

      risc::InstanceTree::Element sub_path;

      sub_path.first = module_instance->get_definition();
      sub_path.second = *iter;

      root.push_back(sub_path);
      get_all_instances_of(instances, root, type, false);
      root.pop_back();
    }

    for(risc::HierarchicalChannelInstanceVectorIter
        iter  = module_instance->get_definition()->sub_hier_channel_.begin();
        iter != module_instance->get_definition()->sub_hier_channel_.end();
        iter++) {

      risc::InstanceTree::Element sub_path;

      sub_path.first = module_instance->get_definition();
      sub_path.second = *iter;

      root.push_back(sub_path);
      get_all_instances_of(instances, root, type, false);
      root.pop_back();
    }

    if(type == risc::Definition::EventType) {
      for(risc::EventVectorIter
          iter  = module_instance->get_definition()->events_.begin();
          iter != module_instance->get_definition()->events_.end();
          iter++) {

        risc::InstanceTree temp = root;
        risc::InstanceTree::Element e(module_instance->get_definition(), *iter);
        temp.push_back(e);
        instances.push_back(temp);
      }
    }

    if(type == risc::Definition::ModuleInstanceType) {

      if(first) {
        instances.push_back(root);
      }

      for(risc::ModuleInstanceVectorIter
          iter  = module_instance->get_definition()->sub_modules_.begin();
          iter != module_instance->get_definition()->sub_modules_.end();
          iter++) {

        risc::InstanceTree temp = root;
        risc::InstanceTree::Element e(module_instance->get_definition(), *iter);
        temp.push_back(e);
        instances.push_back(temp);
      }
    }
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
