#include "systemc.h"

#ifndef WIDTH           // size of the image in pixels
#define WIDTH           640
#endif
#ifndef HEIGHT
#define HEIGHT          512
#endif

typedef struct Image
{
    unsigned char R[HEIGHT][WIDTH];
    unsigned char G[HEIGHT][WIDTH];
    unsigned char B[HEIGHT][WIDTH];

} IMAGE;

typedef struct Coordinates
{
    float_t l;  // left bound, e.g. -2.5
    float_t r;  // right bound, e.g. 1.0
    float_t b;  // bottom bound, e.g. -1.0
    float_t t;  // top bound, e.g. 1.0

} COORDS;

const unsigned char palette[2][3] = {
    //  r    g    b
    {   0,   0,   0 },  // 0, black
    { 255, 255, 255 }}; // 1, white

class Test: public sc_module {

public:

  SC_HAS_PROCESS(Test);

  Test(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(thread);
  }

  void thread()
  {
    int rowStart = 0;
    int rowEnd = 0;
    for(int row = rowStart; row < rowEnd; row++) {

      for(int col = 0; col < WIDTH; col++) {

        float_t x0 = coords.l + (coords.r-coords.l)*((float_t)col/(float_t)WIDTH);
        float_t y0 = coords.b + (coords.t-coords.b)*((float_t)row/(float_t)HEIGHT);
        float_t x = 0;
        float_t y = 0;
        unsigned int iteration = 0;
        const unsigned int max_iteration = 4096;

        while ( x*x + y*y < 2*2  &&  iteration < max_iteration ) {
          float_t xtemp = x*x - y*y + x0;
          y = 2*x*y + y0;
          x = xtemp;
          iteration = iteration + 1;
        }

        unsigned char color = iteration % 6;
        image.R[HEIGHT-1-row][col] = palette[color][0];
        image.G[HEIGHT-1-row][col] = palette[color][1];
        image.B[HEIGHT-1-row][col] = palette[color][2];
      }
    }
  }

  COORDS coords;
  IMAGE image;
};

int sc_main(int argc, char **argv)
{
  Test t("test");
  return 0;
}
