#include "dyn_analysis_instr.h"

#include <string>

#include "../tools/c_definitions.h"
#include "../tools/systemc_definitions.h"
#include "../tools/variable_helper.h"

#include "../internal_representation/design.h"
#include "../internal_representation/event.h"
#include "../internal_representation/event_and_list.h"
#include "../internal_representation/event_or_list.h"
#include "../internal_representation/hierarchical_channel.h"
#include "../internal_representation/hierarchical_channel_instance.h"
#include "../internal_representation/inoutport.h"
#include "../internal_representation/inport.h"
#include "../internal_representation/instance.h"
#include "../internal_representation/instance.h"
#include "../internal_representation/module.h"
#include "../internal_representation/outport.h"
#include "../internal_representation/primitive_channel_instance.h"
#include "../internal_representation/variable.h"

std::pair<SgForStatement*, SgVariableDeclaration*>
risc::inst::insert_loop_iteration_over_vector_elements(
    SgVariableDeclaration *systemc_event_list,
    SgBasicBlock *body)
{
  // The class sc_event_and_list / sc_event_or_list is derived from the class
  // sc_event_list which has the member m_events for the registered events.
  // So, we have to get the base class first. Afterwards, we can get the
  // declaration of the member variable m_events
  SgClassDeclaration *event_list_type_decl
    = isSgClassDeclaration(isSgClassType(
          systemc_event_list->get_decl_item(0)->get_type())->get_declaration());
  event_list_type_decl
    = isSgClassDeclaration(event_list_type_decl->get_definingDeclaration());

  SgClassDeclaration *sc_event_list_class_decl
    = event_list_type_decl->get_definition()->get_inheritances().at(0)
        ->get_base_class();
  sc_event_list_class_decl =
    isSgClassDeclaration(sc_event_list_class_decl->get_definingDeclaration());

  SgVariableDeclaration *event_list = NULL;

  for(SgDeclarationStatementPtrList::iterator
      decl_iter
         = sc_event_list_class_decl->get_definition()->get_members().begin();
      decl_iter
        != sc_event_list_class_decl->get_definition()->get_members().end();
      decl_iter++) {

    if(SageInterface::get_name(*decl_iter)
        == "_variable_declaration_m_events") {

      event_list = isSgVariableDeclaration(*decl_iter);
    }
  }

  // building a reference expression of the event variable
  SgVarRefExp *dot_first = SageBuilder::buildVarRefExp(systemc_event_list);
  SgVarRefExp *dot_second = SageBuilder::buildVarRefExp(event_list);
  SgDotExp *dot_expr = SageBuilder::buildDotExp(dot_first,dot_second);

  event_list_type_decl = isSgClassDeclaration(isSgClassType(
          event_list->get_decl_item(0)->get_type())->get_declaration());
  event_list_type_decl
    = isSgClassDeclaration(event_list_type_decl->get_definingDeclaration());

  // These are the member functions/types of the event vector
  SgMemberFunctionDeclaration *begin = NULL;
  SgMemberFunctionDeclaration *end = NULL;
  SgTemplateInstantiationDecl *const_iterator = NULL;

  for(SgDeclarationStatementPtrList::iterator
      decl_iter
         = event_list_type_decl->get_definition()->get_members().begin();
      decl_iter
        != event_list_type_decl->get_definition()->get_members().end();
      decl_iter++) {

    if(SageInterface::get_name(*decl_iter) == "const_iterator") {

      const_iterator
        = isSgTemplateInstantiationDecl(
            isSgTypedefDeclaration(*decl_iter)->get_declaration());
      break;
    }
  }

  for(SgDeclarationStatementPtrList::iterator
      func_decl_iter
         = event_list_type_decl->get_definition()->get_members().begin();
      func_decl_iter
        != event_list_type_decl->get_definition()->get_members().end();
      func_decl_iter++) {

    SgMemberFunctionDeclaration *func_decl
      = isSgMemberFunctionDeclaration(*func_decl_iter);
    if(func_decl == NULL) {
      continue;
    }

    if(SageInterface::get_name(*func_decl_iter) == "begin") {

      if(const_iterator->get_type() ==
          isSgTypedefType(func_decl->get_orig_return_type())->get_base_type()) {
        begin = func_decl;
      }
    }

    if(SageInterface::get_name(*func_decl_iter) == "end") {

      if(const_iterator->get_type() ==
          isSgTypedefType(func_decl->get_orig_return_type())->get_base_type()) {
        end = func_decl;
      }
    }
  }

  // SgForInitStatement
  SgDotExp *dot_exp =
  SageBuilder::buildDotExp(dot_expr,
      SageBuilder::buildMemberFunctionRefExp(
          isSgMemberFunctionSymbol(
              begin->search_for_symbol_from_symbol_table()), false, false));

  SgFunctionCallExp *call_exp = SageBuilder::buildFunctionCallExp(dot_exp);

  SgAssignInitializer *assign_init
    = SageBuilder::buildAssignInitializer(call_exp);

  SgVariableDeclaration *var_decl_init = NULL;

  var_decl_init = SageBuilder::buildVariableDeclaration(
                      "iter_" + SageInterface::get_name(systemc_event_list),
                      const_iterator->get_type(),
                      assign_init);

  // Creating the for loop
  SgForInitStatement* for_init_stmt = SageBuilder::buildForInitStatement();

  // Adding the declaration of the iterator the the function body
  SageInterface::prependStatement(var_decl_init, body);


  // ForTestStatement
  SgVarRefExp *dot_first1 = SageBuilder::buildVarRefExp(systemc_event_list);
  SgVarRefExp *dot_second1 = SageBuilder::buildVarRefExp(event_list);
  SgDotExp *dot_expr1 = SageBuilder::buildDotExp(dot_first1,dot_second1);

  SgDotExp *dot_exp2 =
  SageBuilder::buildDotExp(dot_expr1,
      SageBuilder::buildMemberFunctionRefExp(
                       isSgMemberFunctionSymbol(
                           end->search_for_symbol_from_symbol_table()),
                           false,
                           false));

  SgFunctionCallExp *call_exp2 = SageBuilder::buildFunctionCallExp(dot_exp2);

  SgExpression *test
    = SageBuilder::buildNotEqualOp(
          SageBuilder::buildVarRefExp(
              isSgVariableSymbol(
                  var_decl_init->get_decl_item(0)
                    ->search_for_symbol_from_symbol_table())),
          call_exp2);

  SgExprStatement *test_stmt = SageBuilder::buildExprStatement(test);

  var_decl_init->get_decl_item(0)->search_for_symbol_from_symbol_table();

  // ForIncrement
  SgExpression *increment =
  SageBuilder::buildPlusPlusOp(
      SageBuilder::buildVarRefExp(
          isSgVariableSymbol(
            var_decl_init->get_decl_item(0)
              ->search_for_symbol_from_symbol_table())),
      SgUnaryOp::postfix);

  // Basic Block
  SgBasicBlock *bb = SageBuilder::buildBasicBlock();

  // For statment
  SgForStatement *for_stmt
    = SageBuilder::buildForStatement(for_init_stmt, test_stmt, increment, bb);

  SageInterface::insertStatement(var_decl_init, for_stmt, false);

  return std::pair<SgForStatement*, SgVariableDeclaration*>(
             for_stmt, var_decl_init);
}

void
risc::inst::print_event_list(
    SgFunctionDefinition *func_def,
    SgVariableDeclaration *event_list,
    SgVariableDeclaration *PLACEHOLDER,
    bool is_and_event_list)
{
  std::pair<SgForStatement*, SgVariableDeclaration*> result
    = insert_loop_iteration_over_vector_elements(
        event_list, func_def->get_body());

  // building a reference expression of the event list variable
  SgVarRefExp *event_list_ref_exp
    = SageBuilder::buildVarRefExp(
        isSgVariableSymbol(event_list->get_decl_item(0)
          ->search_for_symbol_from_symbol_table()));

  // building a reference expression of the event variable
  SgVarRefExp *event_ref_exp
    = SageBuilder::buildVarRefExp(
        isSgVariableSymbol(PLACEHOLDER->get_decl_item(0)
          ->search_for_symbol_from_symbol_table()));

  // building variable reference for fprintf(file, ...)
  SgInitializedName *file_param =
    func_def->get_declaration()->get_parameterList()->get_args().at(0);
  SgVarRefExp *file_expr = SageBuilder::buildVarRefExp(
       isSgVariableSymbol(file_param->search_for_symbol_from_symbol_table()));

  // preparing the format string for the fprintf call
  std::string string_value;
  if(is_and_event_list) {
    string_value = "%s<%p{%p,";
  } else {
    string_value = "%s>%p}%p,";
  }
  SgStringVal *format_string = SageBuilder::buildStringVal(string_value);

  SgStringVal *event_list_name
    = SageBuilder::buildStringVal(event_list_ref_exp->unparseToString());

  SgAddressOfOp *address_of_event_list_expr
    = SageBuilder::buildAddressOfOp(event_list_ref_exp);

  SgVarRefExp *iterator_var_ref = SageBuilder::buildVarRefExp(result.second);

  SgPointerDerefExp *address_of_event
    = SageBuilder::buildPointerDerefExp(iterator_var_ref);

  SgExprListExp *expr_list
    = SageBuilder::buildExprListExp(
        file_expr,
        format_string,
        event_list_name,
        address_of_event_list_expr,
        address_of_event);

  SgFunctionSymbol *func_symbol
    = isSgFunctionSymbol(risc::tools::CDefinitions::get_fprintf()
        ->search_for_symbol_from_symbol_table());

  // building the function call
  SgFunctionCallExp *func_call
    = SageBuilder::buildFunctionCallExp(func_symbol,expr_list);

  // building the statement expression for the function call expression
  SgExprStatement *expr_stmt = SageBuilder::buildExprStatement(func_call);

  // inserting the function call
  SageInterface::appendStatement(
                     expr_stmt,
                     isSgBasicBlock((result.first)->get_loop_body()));
}

void
risc::inst::print_variable(
    SgFunctionDefinition *func_def,
    SgVariableDeclaration *var_decl,
    std::string printf_format_string,
    std::string printf_pre_format_string = "")
{
  // building a reference expression of the channel variable
  SgVarRefExp *var_ref
    = SageBuilder::buildVarRefExp(
        isSgVariableSymbol(var_decl->get_decl_item(0)
          ->search_for_symbol_from_symbol_table()));

  std::string string_value(printf_pre_format_string);
  string_value += var_ref->unparseToString();
  string_value += printf_format_string;

  // building variable reference for fprintf(file, ...)
  SgInitializedName *file_param =
    func_def->get_declaration()->get_parameterList()->get_args().at(0);
  SgVarRefExp *file_expr = SageBuilder::buildVarRefExp(
       isSgVariableSymbol(file_param->search_for_symbol_from_symbol_table()));

  // preparing the format string for the fprintf call
  SgStringVal *string_val = SageBuilder::buildStringVal(string_value);
  SgAddressOfOp *address_of_op = SageBuilder::buildAddressOfOp(var_ref);
  SgExprListExp *expr_list
    = SageBuilder::buildExprListExp(file_expr, string_val,address_of_op);

  SgFunctionSymbol *func_symbol
    = isSgFunctionSymbol(risc::tools::CDefinitions::get_fprintf()
        ->search_for_symbol_from_symbol_table());

  // building the function call
  SgFunctionCallExp *func_call
    = SageBuilder::buildFunctionCallExp(func_symbol,expr_list);

  // building the statement expression for the function call expression
  SgExprStatement *expr_stmt = SageBuilder::buildExprStatement(func_call);

  // inserting the function call
  SageInterface::appendStatement(expr_stmt, func_def->get_body());
}

void
risc::inst::print_port(
    SgFunctionDefinition *func_def,
    SgVariableDeclaration *var_decl)
{
  // goal:
  // fprintf(file, "%s:%p", "PORTNAME", &port);

  // building variable reference for file
  SgInitializedName *file_param =
    func_def->get_declaration()->get_parameterList()->get_args().at(0);
  SgVarRefExp *file_var_ref = SageBuilder::buildVarRefExp(
       isSgVariableSymbol(file_param->search_for_symbol_from_symbol_table()));

  // preparing the format string for the fprintf call
  SgStringVal *format_string = SageBuilder::buildStringVal("%s:%p,");

  // building a reference expression of the port variable
  SgVarRefExp *port_var_ref
    = SageBuilder::buildVarRefExp(
        isSgVariableSymbol(var_decl->get_decl_item(0)
          ->search_for_symbol_from_symbol_table()));

  // building string variable of port name
  SgStringVal *port_name_string
    = SageBuilder::buildStringVal(port_var_ref->unparseToString());

  // building function call to get interface of the port
  SgDotExp *arrow_get_curr_proc
    = SageBuilder::buildDotExp(
        port_var_ref,
        SageBuilder::buildMemberFunctionRefExp(
          isSgMemberFunctionSymbol(
            risc::tools::SystemCDefinitions::get_interface()
              ->get_symbol_from_symbol_table()),
          true,
          false));
  SgFunctionCallExp *get_curr_proc
    = SageBuilder::buildFunctionCallExp(arrow_get_curr_proc);

  // parameters for fprintf(...)
  SgExprListExp *expr_list
    = SageBuilder::buildExprListExp(
                       file_var_ref,
                       format_string,
                       port_name_string,
                       get_curr_proc);

  SgFunctionSymbol *func_symbol
    = isSgFunctionSymbol(
        risc::tools::CDefinitions::get_fprintf()
          ->search_for_symbol_from_symbol_table());

  // building the function call
  SgFunctionCallExp *func_call
    = SageBuilder::buildFunctionCallExp(func_symbol,expr_list);

  // building the statement expression for the function call expression
  SgExprStatement *expr_stmt
    = SageBuilder::buildExprStatement(func_call);

  // inserting the function call
  SageInterface::appendStatement(expr_stmt, func_def->get_body());
}

SgFunctionDefinition*
risc::inst::add_function_to_class(
    SgScopeStatement *class_def,
    std::string func_name)
{
  // building the return type of the function
  SgType *func_return_type = SageBuilder::buildVoidType();
  SgName name = func_name;

  SgFunctionParameterList *param_list
    = SageBuilder::buildFunctionParameterList();

  // building the parameter list of the function
  SgInitializedName *parameter =
    SageBuilder::buildInitializedName(
        "file",
        SageBuilder::buildPointerType(
            risc::tools::CDefinitions::get_file()->get_type()));

  SageInterface::appendArg(param_list, parameter);

  // building the function declaration
  SgFunctionDeclaration *mem_func_decl
    = SageBuilder::buildDefiningFunctionDeclaration(name,
          func_return_type, param_list, class_def);

  SageInterface::appendStatement(mem_func_decl, class_def);

  return isSgFunctionDefinition(mem_func_decl->get_definition());
}

void
risc::inst::print_channel(
    SgFunctionDefinition *func_def,
    risc::Instance *channel)
{
  // Parameter file for the fprintf statement
  SgInitializedName *file_param =
    func_def->get_declaration()->get_parameterList()->get_args().at(0);
  SgVarRefExp *file_expr = SageBuilder::buildVarRefExp(
       isSgVariableSymbol(file_param->search_for_symbol_from_symbol_table()));

  // building a reference expression of the channel variable
  SgVarRefExp *var_ref
    = SageBuilder::buildVarRefExp(
        isSgVariableSymbol(channel->get_ast_node()->get_declaration()
          ->get_decl_item(0)->search_for_symbol_from_symbol_table()));

  // getting the mangled name of the channel type
  std::string channel_type_name
    = channel->get_ast_node()->get_type()->get_mangled();

  // preparing the fprintf format string
  std::string string_value = var_ref->unparseToString();
  string_value += "%s%p&" + channel_type_name + ",";
  SgStringVal *string_val = SageBuilder::buildStringVal(string_value);

  SgStringVal *string_val_per
    = string_val_per = SageBuilder::buildStringVal("%");

  SgExprListExp *expr_list
    = SageBuilder::buildExprListExp(file_expr, string_val);

  expr_list->append_expression(string_val_per);

  SgAddressOfOp *address_of_op = SageBuilder::buildAddressOfOp(var_ref);

  SgClassDeclaration *interface
    = tools::SystemCDefinitions::get_sc_interface()->get_declaration();

  SgCastExp *cast = SageBuilder::buildCastExp(
                                     address_of_op,
                                     SageBuilder::buildPointerType(
                                                      interface->get_type()));

  expr_list->append_expression(cast);

  SgFunctionSymbol *func_symbol
    = isSgFunctionSymbol(risc::tools::CDefinitions::get_fprintf()
        ->search_for_symbol_from_symbol_table());

  // building the function call
  SgFunctionCallExp *func_call
    = SageBuilder::buildFunctionCallExp(func_symbol,expr_list);

  // building the statement expression for the function call expression
  SgExprStatement *expr_stmt = SageBuilder::buildExprStatement(func_call);

  // inserting the function call
  SageInterface::appendStatement(expr_stmt, func_def->get_body());
}

void
risc::inst::instrument_reference_printing(std::string filename)
{
  SgFunctionDefinition *sc_main_def
    = tools::SystemCDefinitions::get_sc_main();

  // inserting the required header for the type FILE
  SageInterface::insertHeader(
      SageInterface::getEnclosingSourceFile(sc_main_def),
      "stdio.h",
      true,
      PreprocessingInfo::before);

  // Writing the filename for the dynamic analysis file (DIR) in a variable.
  // The dynamic SystemC library will read this variable and create a file
  // with this name
  SgStringVal *filename_string = SageBuilder::buildStringVal(filename);

  SgVariableDeclaration *var_decl =
    SageBuilder::buildVariableDeclaration(
        "__risc__dir_filename",
        risc::tools::CDefinitions::get_std_string()->get_type(),
        SageBuilder::buildAssignInitializer(filename_string));

  SageInterface::insertStatement(
      sc_main_def->get_declaration()->get_definingDeclaration(),
      var_decl,
      true);

  // Here we are instrumenting the printing of all global variables.
  // The instrumentation is in the function print_global_variables
  for(int file_id = 0;
      file_id < risc::get_project()->numberOfFiles();
      file_id++) {

    SgSourceFile *file
      = isSgSourceFile(&risc::get_project()->get_file(file_id));

    SgFunctionDefinition *print_global_variables =
      add_function_to_class(file->get_globalScope(), "print_global_variables");

    for(SgDeclarationStatementPtrList::iterator
        iter  = file->get_globalScope()->get_declarations().begin();
        iter != file->get_globalScope()->get_declarations().end();
        iter++) {

      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);

      if(var_decl) {

        if(isSgLocatedNode(var_decl)) {

          if(var_decl->get_file_info()->get_file_id() == file_id) {

            print_variable(print_global_variables, var_decl, "-%p,","-");
          }
        }
      }
    }
  }

  // Here, we are instrumenting all module definitions.
  // We are adding the functions print_variables, print_ports, and
  // print_channels which print the associated variables with their addresses.
  for(risc::ModuleVectorIter
      iter_module  = risc::get_design()->module_definitions_.begin();
      iter_module != risc::get_design()->module_definitions_.end();
      iter_module++) {

    SgFunctionDefinition *print_variables =
      add_function_to_class((*iter_module)->get_ast_node(), "print_variables");
    SgFunctionDefinition *print_ports =
      add_function_to_class((*iter_module)->get_ast_node(), "print_ports");
    SgFunctionDefinition *print_channels =
      add_function_to_class((*iter_module)->get_ast_node(), "print_channels");
    SgFunctionDefinition *print_events =
      add_function_to_class((*iter_module)->get_ast_node(), "print_events");

    for(risc::InPortVector::iterator
        iter_port  = (*iter_module)->inports_.begin();
        iter_port != (*iter_module)->inports_.end();
        iter_port++) {
      print_port(print_ports, (*iter_port)->get_ast_node()->get_declaration());
    }
    for(risc::OutPortVector::iterator
        iter_port  = (*iter_module)->outports_.begin();
        iter_port != (*iter_module)->outports_.end();
        iter_port++) {
      print_port(print_ports, (*iter_port)->get_ast_node()->get_declaration());
    }
    for(risc::InOutPortVector::iterator
        iter_port  = (*iter_module)->inoutports_.begin();
        iter_port != (*iter_module)->inoutports_.end();
        iter_port++) {
      print_port(print_ports, (*iter_port)->get_ast_node()->get_declaration());
    }

    std::list<risc::Variable*> references;

    for(risc::VariableVectorIter
        iter_variable  = (*iter_module)->variables_.begin();
        iter_variable != (*iter_module)->variables_.end();
        iter_variable++) {

      SgVariableDeclaration *var_decl
        = (*iter_variable)->get_ast_node()->get_declaration();

      SgVariableSymbol *var_symbol
        = isSgVariableSymbol(
              var_decl->get_decl_item(0)
                ->search_for_symbol_from_symbol_table());

      SgVariableDeclaration *var_decl_2
        = (*iter_variable)->get_ast_node()->get_declaration();
      if(risc::tools::is_reference_variable(var_symbol)) {
        print_variable(print_variables, var_decl_2, "@%p,");
      } else {
        print_variable(print_variables, var_decl_2, "$%p,");
      }
    }

    SgVariableDeclaration *event = NULL;
    for(risc::EventVectorIter
        iter_event  = (*iter_module)->events_.begin();
        iter_event != (*iter_module)->events_.end();
        iter_event++) {

      print_variable(print_events,
          (*iter_event)->get_ast_node()->get_declaration(), "+%p,");
      event = (*iter_event)->get_ast_node()->get_declaration();
    }

    for(risc::EventAndListVectorIter
        iter_event_and_list  = (*iter_module)->event_and_lists_.begin();
        iter_event_and_list != (*iter_module)->event_and_lists_.end();
        iter_event_and_list++) {

      print_event_list(print_events,
          (*iter_event_and_list)->get_ast_node()->get_declaration(),
          event,
          true);
    }

    for(risc::EventOrListVectorIter
        iter_event_or_list  = (*iter_module)->event_or_lists_.begin();
        iter_event_or_list != (*iter_module)->event_or_lists_.end();
        iter_event_or_list++) {

      print_event_list(print_events,
          (*iter_event_or_list)->get_ast_node()->get_declaration(),
          event,
          false);
    }

    for(risc::HierarchicalChannelInstanceVector::iterator
        iter_channel  = (*iter_module)->sub_hier_channel_.begin();
        iter_channel != (*iter_module)->sub_hier_channel_.end();
        iter_channel++) {

      print_channel(print_channels, *iter_channel);
    }

    for(risc::PrimitiveChannelInstanceVector::iterator
        iter_channel  = (*iter_module)->sub_prim_channel_.begin();
        iter_channel != (*iter_module)->sub_prim_channel_.end();
        iter_channel++) {

      print_channel(print_channels, *iter_channel);
    }
  }

  // Same instrumentation for the hierarchical channels like for the modules
  for(risc::HierarchicalChannelVectorIter
      iter_module  = risc::get_design()->hierarchical_channel_.begin();
      iter_module != risc::get_design()->hierarchical_channel_.end();
      iter_module++) {

    SgFunctionDefinition *print_variables =
      add_function_to_class((*iter_module)->get_ast_node(), "print_variables");
    SgFunctionDefinition *print_ports =
      add_function_to_class((*iter_module)->get_ast_node(), "print_ports");
    SgFunctionDefinition *print_channels =
      add_function_to_class((*iter_module)->get_ast_node(), "print_channels");
    SgFunctionDefinition *print_events =
      add_function_to_class((*iter_module)->get_ast_node(), "print_events");

    for(risc::InPortVector::iterator
        iter_port  = (*iter_module)->inports_.begin();
        iter_port != (*iter_module)->inports_.end();
        iter_port++) {
      print_port(print_ports, (*iter_port)->get_ast_node()->get_declaration());
    }
    for(risc::OutPortVector::iterator
        iter_port  = (*iter_module)->outports_.begin();
        iter_port != (*iter_module)->outports_.end();
        iter_port++) {
      print_port(print_ports, (*iter_port)->get_ast_node()->get_declaration());
    }
    for(risc::InOutPortVector::iterator
        iter_port  = (*iter_module)->inoutports_.begin();
        iter_port != (*iter_module)->inoutports_.end();
        iter_port++) {
      print_port(print_ports, (*iter_port)->get_ast_node()->get_declaration());
    }

    std::list<risc::Variable*> references;

    for(risc::VariableVectorIter
        iter_variable  = (*iter_module)->variables_.begin();
        iter_variable != (*iter_module)->variables_.end();
        iter_variable++) {

      SgVariableDeclaration *var_decl
        = (*iter_variable)->get_ast_node()->get_declaration();

      SgVariableSymbol *var_symbol
        = isSgVariableSymbol(
              var_decl->get_decl_item(0)
                ->search_for_symbol_from_symbol_table());

      SgVariableDeclaration *var_decl_2
        = (*iter_variable)->get_ast_node()->get_declaration();
      if(risc::tools::is_reference_variable(var_symbol)) {
        print_variable(print_variables, var_decl_2, "@%p,");
      } else {
        print_variable(print_variables, var_decl_2, "$%p,");
      }
    }

    for(risc::EventVectorIter
        iter_event  = (*iter_module)->events_.begin();
        iter_event != (*iter_module)->events_.end();
        iter_event++) {

      print_variable(print_events,
          (*iter_event)->get_ast_node()->get_declaration(), "+%p,");
    }

    for(risc::HierarchicalChannelInstanceVector::iterator
        iter_channel  = (*iter_module)->sub_hier_channel_.begin();
        iter_channel != (*iter_module)->sub_hier_channel_.end();
        iter_channel++) {

      print_channel(print_channels, *iter_channel);
    }

    for(risc::PrimitiveChannelInstanceVector::iterator
        iter_channel  = (*iter_module)->sub_prim_channel_.begin();
        iter_channel != (*iter_module)->sub_prim_channel_.end();
        iter_channel++) {

      print_channel(print_channels, *iter_channel);
    }
  }
}

