#ifndef DYN_ANALYSIS_INSTR_H_INCLUDED_
#define DYN_ANALYSIS_INSTR_H_INCLUDED_

#include "rose.h"

#include <list>
#include <utility>


namespace risc
{

class Instance;

namespace inst
{

/**
 * \fn        print_event_list
 * \brief    This function adds the printing of event list and the individual
 *           elements to the function func_def
 */
void
print_event_list(
    SgFunctionDefinition *func_def,
    SgVariableDeclaration *systemc_event_list,
    SgVariableDeclaration *PLACEHOLDER,
    bool is_and_event_list);

/**
 * \fn       print_variable
 * \brief    This function adds the printing of the variable var_decl to the
 *           function func_def
 *           This function is only for internal use.
 */
void
print_variable(
    SgFunctionDefinition *func_def,
    SgVariableDeclaration *var_decl,
    std::string printf_format_string,
    std::string printf_pre_format_string);


/**
 * \fn       print_port
 * \brief    This function adds the printing of the variable var_decl to the
 *           function func_def
 *           This function is only for internal use.
 */
void
print_port(
    SgFunctionDefinition *func_def,
    SgVariableDeclaration *var_decl);


/**
 * \fn       print_channel
 * \brief    This function adds the printing of the channel channel to the
 *           function func_def
 *           This function is only for internal use.
 */
void
print_channel(SgFunctionDefinition *func_def, risc::Instance *channel);

/**
 * \fn       add_function_to_class
 * \brief    This function adds a new function with the name func_name to the
 *           class class_def. The return type will be void and the only
 *           parameter is FILE *file
 *           This function is only for internal use.
 */
SgFunctionDefinition*
add_function_to_class(SgScopeStatement *class_def, std::string func_name);

/**
 * \fn       instrument_reference_printing
 *           The filename will be used for the dir file.
 * \brief    The function should be called for the instrumentation of the
 *           variable address printing of the design
 */
void
instrument_reference_printing(std::string filename);

/**
 * \fn       insert_loop_iteration_over_vector_elements
 * \brief    This function appends a for loop to the given basic block.
 *           The loop body will be empty.
 *           The function returns the for loop statement and the variable
 *           declaration of the iterator variable
 */
std::pair<SgForStatement*, SgVariableDeclaration*>
insert_loop_iteration_over_vector_elements(
    SgVariableDeclaration *event_list,
    SgBasicBlock *body);

}; // end of namespace inst

}; // end of namespace risc

#endif /* DYN_ANALYSIS_INSTR_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
