#ifndef TIME_H_INCLUDED_
#define TIME_H_INCLUDED_

#include <iostream>

namespace risc {

namespace sg {

/**
 * \class Time
 * \brief This class represents timing in discrete event simulation
 */
class Time {
public:

  /**
   * \fn       Time
   * \brief    This constructor initializes the time values with (0,0)
   */
  Time();

  /**
   * \fn       Time
   * \brief    This constructor initializes the time values with user custom
   *           values.
   */
  Time(unsigned long long time_units, int delta_count);

  /**
   * \fn       Time
   * \brief    This constructor initializes the time values with user custom
   *           values.
   */
  Time(unsigned long long time_units, int delta_count, bool infinite);

  /**
   * \fn       Time
   * \brief    This constructor should be used to set the infinite flag
   */
  Time(bool infinite);

  int delta_count_;
  unsigned long long time_units_;

  bool operator<(const Time &ohs);
  bool operator>(const Time &ohs);

  bool operator<=(const Time &ohs);
  bool operator>=(const Time &ohs);

  bool operator==(const Time &ohs);
  bool operator!=(const Time &ohs);

  Time operator+(const Time &ohs);
  Time& operator+=(const Time &ohs);

  /**
   * \brief This flag is set to describe an infinite amount of time
   */
  bool infinite_;

  /**
   * \fn       one_delta_cycle_ahead
   * \brief    The function checks if the difference between this and the given
   *           time stamp is exactly one delta cycle
   */
  bool one_delta_cycle_ahead(const Time &ohs);



  static long long convert_si_magnitude(std::string magnitude);

};

std::ostream& operator<<(std::ostream& os, const risc::sg::Time& time);

}; // end of namesapce sg

}; // end of namespace risc

#endif /* TIME_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
