#ifndef VARIABLE_HELPER_H_INCLUDED_
#define VARIABLE_HELPER_H_INCLUDED_

#include "rose.h"

namespace risc {

namespace tools {

/**
 * \function is_reference_variable
 * \brief    This function return true if the given variable symbol is a
 *           reference
 */
bool is_reference_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_pointer_variable
 * \brief    This function return true if the given variable symbol is a
 *           pointer
 */
bool is_pointer_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_global_variable
 * \brief    This function return true if the given variable symbol is a
 *           global variable
 */
bool is_global_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_local_variable
 * \brief    This function return true if the given variable symbol is a
 *           local variable
 */
bool is_local_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_member_variable
 * \brief    This function return true if the given variable symbol is a
 *           member variable
 */
bool is_member_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_static_variable
 * \brief    This function return true if the given variable symbol is a
 *           static variable
 */
bool is_static_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_namespace_variable
 * \brief    This function return true if the given variable symbol is a
 *           namespace variable
 */
bool is_namespace_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_member_variable
 * \brief    This function return true if the given variable symbol is a
 *           function variable
 */
bool is_function_parameter_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_member_of
 * \brief    This function returns true if a given variable symbol is defined
 *           'in' another variable symbol. For instance 'class Foo { int i;}'
 *           'i' is defined in 'Foo'
 */
bool is_member_of(SgVariableSymbol *is, SgSymbol *of);

/**
 * \function is_loop_variable
 * \brief    This function return true if the given variable symbol is a
 *           loop variable
 */
bool is_loop_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_stack_variable
 * \brief    This function return true if the given variable symbol is a
 *           variable which will be stored on the stack
 */
bool is_stack_variable(SgVariableSymbol *var_symbol);

/**
 * \function is_data_segment_variable
 * \brief    This function return true if the given variable symbol is a
 *           variable which will be stored on the stack
 */
bool is_data_segment_variable(SgVariableSymbol *var_symbol);

/**
 * \function analyze_conflict_free
 * \brief    This function checks if a variable is annoated with a pragma
 *           as conflict free
 */
void analyze_conflict_free(SgVariableSymbol *symbol);

/**
 * \function is_channel_variable
 * \brief    Returns true when the variable is declared in a channel
 */
bool is_channel_variable(SgVariableSymbol *symbol);

}; // end of namespace tools

}; // end of namespace risc

#endif /* VARIABLE_HELPER_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
