#include "systemc.h"

class Sender: public sc_module {
public:
  sc_fifo_out<int> fifo_out1;
  sc_fifo_out<int> fifo_out2;
  sc_fifo_out<int> fifo_out3;

  Sender(sc_module_name mn):
    sc_module(mn)
  { }
};

class Receiver: public sc_module {

public:
  sc_fifo_in<int> fifo_in1;
  sc_fifo_in<int> fifo_in2;
  sc_fifo_in<int> fifo_in3;

  Receiver(sc_module_name mn):
    sc_module(mn)
  { }
};


class ChannelInterface:
  virtual public sc_interface {

public:
  virtual void read(void) = 0;
  virtual void write() = 0;
};

class Channel:
  public sc_channel,
  public ChannelInterface
{
public:
  Channel(sc_module_name name):
    sc_channel(name)
  { }

  void read()
  { }

  void write()
  { }
};

class SenderC: public sc_module {
public:
  sc_port<ChannelInterface> out1;
  sc_port<ChannelInterface> out2;
  sc_port<ChannelInterface> out3;

  SenderC(sc_module_name mn):
    sc_module(mn)
  { }
};

class ReceiverC: public sc_module {

public:
  sc_port<ChannelInterface> in1;
  sc_port<ChannelInterface> in2;
  sc_port<ChannelInterface> in3;

  ReceiverC(sc_module_name mn):
    sc_module(mn)
  { }
};

class Top: public sc_module {

public:
  sc_fifo<int> fifo_beoe11;
  sc_fifo<int> fifo_beoe12;
  sc_fifo<int> fifo_beoe13;
  sc_fifo<int> fifo_beoe21;
  sc_fifo<int> fifo_beoe22;
  sc_fifo<int> fifo_beoe23;
  sc_fifo<int> fifo_beoe31;
  sc_fifo<int> fifo_beoe32;
  sc_fifo<int> fifo_beoe33;

  Sender s_beoe1;
  Sender s_beoe2;
  Sender s_beoe3;

  Receiver r_beoe1;
  Receiver r_beoe2;
  Receiver r_beoe3;

  sc_fifo<int> fifo_ctor11;
  sc_fifo<int> fifo_ctor12;
  sc_fifo<int> fifo_ctor13;
  sc_fifo<int> fifo_ctor21;
  sc_fifo<int> fifo_ctor22;
  sc_fifo<int> fifo_ctor23;
  sc_fifo<int> fifo_ctor31;
  sc_fifo<int> fifo_ctor32;
  sc_fifo<int> fifo_ctor33;

  Sender s_ctor1;
  Sender s_ctor2;
  Sender s_ctor3;

  Receiver r_ctor1;
  Receiver r_ctor2;
  Receiver r_ctor3;

  /////////////////////////////////
  SenderC sc_beoe1;
  SenderC sc_beoe2;
  SenderC sc_beoe3;

  ReceiverC rc_beoe1;
  ReceiverC rc_beoe2;
  ReceiverC rc_beoe3;

  Channel channel_beoe11;
  Channel channel_beoe12;
  Channel channel_beoe13;
  Channel channel_beoe21;
  Channel channel_beoe22;
  Channel channel_beoe23;
  Channel channel_beoe31;
  Channel channel_beoe32;
  Channel channel_beoe33;


  Channel channel_ctor11;
  Channel channel_ctor12;
  Channel channel_ctor13;
  Channel channel_ctor21;
  Channel channel_ctor22;
  Channel channel_ctor23;
  Channel channel_ctor31;
  Channel channel_ctor32;
  Channel channel_ctor33;


  SenderC sc_ctor1;
  SenderC sc_ctor2;
  SenderC sc_ctor3;

  ReceiverC rc_ctor1;
  ReceiverC rc_ctor2;
  ReceiverC rc_ctor3;

  Top(sc_module_name mn):
    sc_module(mn),
    s_beoe1("s_beoe1"),
    s_beoe2("s_beoe2"),
    s_beoe3("s_beoe3"),
    r_beoe1("r_beoe1"),
    r_beoe2("r_beoe2"),
    r_beoe3("r_beoe3"),
    s_ctor1("s_ctor1"),
    s_ctor2("s_ctor2"),
    s_ctor3("s_ctor3"),
    r_ctor1("r_ctor1"),
    r_ctor2("r_ctor2"),
    r_ctor3("r_ctor3"),

    sc_beoe1("sc_beoe1"),
    sc_beoe2("sc_beoe2"),
    sc_beoe3("sc_beoe3"),
    rc_beoe1("rc_beoe1"),
    rc_beoe2("rc_beoe2"),
    rc_beoe3("rc_beoe3"),

    sc_ctor1("sc_ctor1"),
    sc_ctor2("sc_ctor2"),
    sc_ctor3("sc_ctor3"),
    rc_ctor1("rc_ctor1"),
    rc_ctor2("rc_ctor2"),
    rc_ctor3("rc_ctor3"),
    channel_beoe11("channel_beoe11"),
    channel_beoe12("channel_beoe12"),
    channel_beoe13("channel_beoe13"),
    channel_beoe21("channel_beoe21"),
    channel_beoe22("channel_beoe22"),
    channel_beoe23("channel_beoe23"),
    channel_beoe31("channel_beoe31"),
    channel_beoe32("channel_beoe32"),
    channel_beoe33("channel_beoe33"),

    channel_ctor11("channel_ctor11"),
    channel_ctor12("channel_ctor12"),
    channel_ctor13("channel_ctor13"),
    channel_ctor21("channel_ctor21"),
    channel_ctor22("channel_ctor22"),
    channel_ctor23("channel_ctor23"),
    channel_ctor31("channel_ctor31"),
    channel_ctor32("channel_ctor32"),
    channel_ctor33("channel_ctor33")
  {
    s_ctor1.fifo_out1.bind(fifo_ctor11);
    s_ctor1.fifo_out2.bind(fifo_ctor12);
    s_ctor1.fifo_out3.bind(fifo_ctor13);
    r_ctor1.fifo_in1.bind(fifo_ctor11);
    r_ctor1.fifo_in2.bind(fifo_ctor12);
    r_ctor1.fifo_in3.bind(fifo_ctor13);

    s_ctor2.fifo_out1(fifo_ctor21);
    s_ctor2.fifo_out2(fifo_ctor22);
    s_ctor2.fifo_out3(fifo_ctor23);
    r_ctor2.fifo_in1(fifo_ctor21);
    r_ctor2.fifo_in2(fifo_ctor22);
    r_ctor2.fifo_in3(fifo_ctor23);

    s_ctor3(fifo_ctor31, fifo_ctor32, fifo_ctor33);
    r_ctor3(fifo_ctor31, fifo_ctor32, fifo_ctor33);


    sc_ctor1.out1.bind(channel_ctor11);
    sc_ctor1.out2.bind(channel_ctor12);
    sc_ctor1.out3.bind(channel_ctor13);
    rc_ctor1.in1.bind(channel_ctor11);
    rc_ctor1.in2.bind(channel_ctor12);
    rc_ctor1.in3.bind(channel_ctor13);

    sc_ctor2.out1(channel_ctor21);
    sc_ctor2.out2(channel_ctor22);
    sc_ctor2.out3(channel_ctor23);
    rc_ctor2.in1(channel_ctor21);
    rc_ctor2.in2(channel_ctor22);
    rc_ctor2.in3(channel_ctor23);

    sc_ctor3(channel_ctor31, channel_ctor32, channel_ctor33);
    rc_ctor3(channel_ctor31, channel_ctor32, channel_ctor33);
  }

  void before_end_of_elaboration()
  {
    s_beoe1.fifo_out1.bind(fifo_beoe11);
    s_beoe1.fifo_out2.bind(fifo_beoe12);
    s_beoe1.fifo_out3.bind(fifo_beoe13);
    r_beoe1.fifo_in1.bind(fifo_beoe11);
    r_beoe1.fifo_in2.bind(fifo_beoe12);
    r_beoe1.fifo_in3.bind(fifo_beoe13);

    s_beoe2.fifo_out1(fifo_beoe21);
    s_beoe2.fifo_out2(fifo_beoe22);
    s_beoe2.fifo_out3(fifo_beoe23);
    r_beoe2.fifo_in1(fifo_beoe21);
    r_beoe2.fifo_in2(fifo_beoe22);
    r_beoe2.fifo_in3(fifo_beoe23);

    s_beoe3(fifo_beoe31, fifo_beoe32, fifo_beoe33);
    r_beoe3(fifo_beoe31, fifo_beoe32, fifo_beoe33);


    sc_beoe1.out1.bind(channel_beoe11);
    sc_beoe1.out2.bind(channel_beoe12);
    sc_beoe1.out3.bind(channel_beoe13);
    rc_beoe1.in1.bind(channel_beoe11);
    rc_beoe1.in2.bind(channel_beoe12);
    rc_beoe1.in3.bind(channel_beoe13);

    sc_beoe2.out1(channel_beoe21);
    sc_beoe2.out2(channel_beoe22);
    sc_beoe2.out3(channel_beoe23);
    rc_beoe2.in1(channel_beoe21);
    rc_beoe2.in2(channel_beoe22);
    rc_beoe2.in3(channel_beoe23);

    sc_beoe3(channel_beoe31, channel_beoe32, channel_beoe33);
    rc_beoe3(channel_beoe31, channel_beoe32, channel_beoe33);
  }
};

int sc_main(int argc, char **argv)
{
  Top t("top");
  sc_start();
  return 0;
}
