#include <iostream>

#include "risc.h"

int main(int argc, char* argv[])
{
  std::cout << std::endl
            << "start processing file: " __FILE__
            << " ..."
            << std::endl;

  risc::Design d(argc,argv);

  std::cout << "+++++++++++++++++++ Module Definitions +++++++++++++++++++"
            << std::endl;

  // iterating all modules in the design
  for(risc::ModuleVectorIter
      iter  = d.module_definitions_.begin();
      iter != d.module_definitions_.end(); iter++) {

    // print all sub modules in the individual module definitions
    (*iter)->print_module_definition();

    std::cout << std::endl;
  }
#if 0
  std::cout << "+++++++++++++++++++ Top Modules ++++++++++++++++++++++++++"
            << std::endl;

  // iterating all modules in the design
  for(risc::InstanceVectorIter
      iter  = d.top_modules_.begin();
      iter != d.top_modules_.end(); iter++) {

    std::cout <<  (*iter)->get_ast_type_name()
              << " "
              << (*iter)->get_name()
              << std::endl;
  }


  std::cout << "+++++++++++++++++++ Hierarchy Modules ++++++++++++++++++++"
            << std::endl;

  for(risc::InstanceVectorIter
      iter  = d.top_modules_.begin();
      iter != d.top_modules_.end(); iter++) {

    std::cout <<  (*iter)->get_ast_type_name()
              << " "
              << (*iter)->get_name()
              << std::endl;

    risc::ModuleInstance *module = dynamic_cast<risc::ModuleInstance*>(*iter);

    module->get_definition()->print_hierarchy(1);
  }


  std::cout << "+++++++++++++++++++ Building Segment Graph ++++++++++++++++++++"
            << std::endl;

  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
    risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    std::cout << "Module: " <<  (*iter_module)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;

      sg.add_function(*iter_thread);
    }
  }

  sg.clean_graph();
  sg.print_graph("static_mapping_sg.dot");
#endif
  std::cout << "end processing file: " __FILE__
            << " ..."
            << std::endl
            << std::endl;

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
