// player_duplicated_channels.cc: derived from examples/demo/play.cpp
// 07/10/17 RD  split the delay into two parts (as suggested by ZC)
// 04/28/17 RD   all instances are duplicated here: one instance per class
// 04/13/17 RD   replaced the broken channel, updated to latest play.cpp
// 04/07/17 ZC   initial version (based on conceptional DVD player play.cpp)

#include <stdio.h>
#include "systemc.h"

// frame rates
#define AUDIO_DELAY     26120, SC_US
#define VIDEO_DELAY     33330, SC_US
#define HALF_AUDIO_DELAY 26120/2, SC_US
#define HALF_VIDEO_DELAY 33330/2, SC_US

// simulated time
//#define CLIP_LENGTH	10, SC_SEC	// original model
#define CLIP_LENGTH	 1, SC_SEC	// shorter regression test

// simulated workload unit
#define AUDIOLOAD        7500000
#define VIDEOLOAD       10000000

// illustrate and check frame decoding
#define PRINT_TIME
#define PRINT_FRAMES
#define COUNT_FRAMES

// buffer size in queue channels
#define BUFSIZE		10ul

// queue communication

class i_long_long_senderV: virtual public sc_interface
{
  public:
    virtual void send(long long d) = 0;
};

class i_long_long_receiverV: virtual public sc_interface
{
  public:
    virtual long long receive(void) = 0;
};

class c_long_long_queueV: public sc_channel,
        public i_long_long_senderV, public i_long_long_receiverV
{
  public:
    c_long_long_queueV(sc_module_name name, unsigned long size):
                sc_channel(name), m_size(size)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      buffer = new long long[size];
    }

    ~c_long_long_queueV(void)
    {
      delete[] buffer;
    }

    long long receive(void)
    {
      long long d;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      return d;
    }

    void send(long long d)
    {
      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    long long      *buffer;
    unsigned long  m_size;
};
class i_long_long_senderVS: virtual public sc_interface
{
  public:
    virtual void send(long long d) = 0;
};

class i_long_long_receiverVS: virtual public sc_interface
{
  public:
    virtual long long receive(void) = 0;
};

class c_long_long_queueVS: public sc_channel,
        public i_long_long_senderVS, public i_long_long_receiverVS
{
  public:
    c_long_long_queueVS(sc_module_name name, unsigned long size):
                sc_channel(name), m_size(size)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      buffer = new long long[size];
    }

    ~c_long_long_queueVS(void)
    {
      delete[] buffer;
    }

    long long receive(void)
    {
      long long d;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      return d;
    }

    void send(long long d)
    {
      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    long long      *buffer;
    unsigned long  m_size;
};


class i_long_long_senderL: virtual public sc_interface //L for left
{
  public:
    virtual void send(long long d) = 0;
};

class i_long_long_receiverL: virtual public sc_interface
{
  public:
    virtual long long receive(void) = 0;
};

class c_long_long_queueL: public sc_channel,
        public i_long_long_senderL, public i_long_long_receiverL
{
  public:
    c_long_long_queueL(sc_module_name name, unsigned long size):
                sc_channel(name), m_size(size)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      buffer = new long long[size];
    }

    ~c_long_long_queueL(void)
    {
      delete[] buffer;
    }

    long long receive(void)
    {
      long long d;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      return d;
    }

    void send(long long d)
    {
      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    long long      *buffer;
    unsigned long  m_size;
};
class i_long_long_senderLS: virtual public sc_interface //L for left
{
  public:
    virtual void send(long long d) = 0;
};

class i_long_long_receiverLS: virtual public sc_interface
{
  public:
    virtual long long receive(void) = 0;
};

class c_long_long_queueLS: public sc_channel,
        public i_long_long_senderLS, public i_long_long_receiverLS
{
  public:
    c_long_long_queueLS(sc_module_name name, unsigned long size):
                sc_channel(name), m_size(size)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      buffer = new long long[size];
    }

    ~c_long_long_queueLS(void)
    {
      delete[] buffer;
    }

    long long receive(void)
    {
      long long d;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      return d;
    }

    void send(long long d)
    {
      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    long long      *buffer;
    unsigned long  m_size;
};


class i_long_long_senderR: virtual public sc_interface
{
  public:
    virtual void send(long long d) = 0;
};

class i_long_long_receiverR: virtual public sc_interface
{
  public:
    virtual long long receive(void) = 0;
};

class c_long_long_queueR: public sc_channel,
        public i_long_long_senderR, public i_long_long_receiverR
{
  public:
    c_long_long_queueR(sc_module_name name, unsigned long size):
                sc_channel(name), m_size(size)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      buffer = new long long[size];
    }

    ~c_long_long_queueR(void)
    {
      delete[] buffer;
    }

    long long receive(void)
    {
      long long d;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      return d;
    }

    void send(long long d)
    {
      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    long long      *buffer;
    unsigned long  m_size;
};
class i_long_long_senderRS: virtual public sc_interface
{
  public:
    virtual void send(long long d) = 0;
};

class i_long_long_receiverRS: virtual public sc_interface
{
  public:
    virtual long long receive(void) = 0;
};

class c_long_long_queueRS: public sc_channel,
        public i_long_long_senderRS, public i_long_long_receiverRS
{
  public:
    c_long_long_queueRS(sc_module_name name, unsigned long size):
                sc_channel(name), m_size(size)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      buffer = new long long[size];
    }

    ~c_long_long_queueRS(void)
    {
      delete[] buffer;
    }

    long long receive(void)
    {
      long long d;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      return d;
    }

    void send(long long d)
    {
      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    long long      *buffer;
    unsigned long  m_size;
};

SC_MODULE(Stimulus)
{
  sc_port<i_long_long_senderVS> VideoStream;
  sc_port<i_long_long_senderLS> AudioStreamL;
  sc_port<i_long_long_senderRS> AudioStreamR;

  void main(void)
  {
    sc_time     t1(0, SC_US);
    sc_time     t2(0, SC_US);
    long long   video = 42;
    long long   audioLeft = 27;
    long long   audioRight = 15;

    while ((t1 < sc_time(CLIP_LENGTH)) || (t2 < sc_time(CLIP_LENGTH)))
    {
      if (t1 < t2)
      {
        VideoStream->send(video);
        t1 += sc_time(VIDEO_DELAY);
      }
      else
      {
        AudioStreamL->send(audioLeft);
        AudioStreamR->send(audioRight);
        t2 += sc_time(AUDIO_DELAY);
      }
    }
  }

  SC_CTOR(Stimulus)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(VideoCodec)
{
  sc_port<i_long_long_receiverVS> p1;
  sc_port<i_long_long_senderV>   p2;

  long long d;

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < VIDEOLOAD; i ++)
    {
      d *= x;
    }
    return d;
  }

  void main(void)
  {
    long long inFrame, outFrame;

    d = 1;

    while(1)
    {
      inFrame = p1->receive();
      wait(HALF_VIDEO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_VIDEO_DELAY);
      p2->send(outFrame);
    }
  }

  SC_CTOR(VideoCodec)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(AudioCodecL)
{
  sc_port<i_long_long_receiverLS> p1;
  sc_port<i_long_long_senderL>   p2;

  long long d;

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < AUDIOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }

  void main(void)
  {
    long long inFrame, outFrame;

    d = 1;

    while(1)
    {
      inFrame = p1->receive();
      wait(HALF_AUDIO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_AUDIO_DELAY);
      p2->send(outFrame);
    }
  }

  SC_CTOR(AudioCodecL)
  {
    SC_THREAD(main);
  }
};


SC_MODULE(AudioCodecR)
{
  sc_port<i_long_long_receiverRS> p1;
  sc_port<i_long_long_senderR>   p2;

  long long d;

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < AUDIOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }

  void main(void)
  {
    long long inFrame, outFrame;

    d = 1;

    while(1)
    {
      inFrame = p1->receive();
      wait(HALF_AUDIO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_AUDIO_DELAY);
      p2->send(outFrame);
    }
  }

  SC_CTOR(AudioCodecR)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(DUT)
{
  sc_port<i_long_long_receiverVS> VideoIn;
  sc_port<i_long_long_receiverLS> AudioInL;
  sc_port<i_long_long_receiverRS> AudioInR;
  sc_port<i_long_long_senderV>   VideoOut;
  sc_port<i_long_long_senderL>   AudioOutLeft;
  sc_port<i_long_long_senderR>   AudioOutRight;

  VideoCodec video;
  AudioCodecL left;
  AudioCodecR right;

  void before_end_of_elaboration()
  {
    video.p1.bind(VideoIn);
    video.p2.bind(VideoOut);

    left.p1.bind(AudioInL);
    left.p2.bind(AudioOutLeft);

    right.p1.bind(AudioInR);
    right.p2.bind(AudioOutRight);
  }

  SC_CTOR(DUT): video("video"), left("left"), right("right")
  { }
};

SC_MODULE(Display)
{
  sc_port<i_long_long_receiverV> VideoIn;
#ifdef COUNT_FRAMES
  unsigned FramesDisplayed;
#endif

  void display(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    putchar('V');       // display the frame
    fflush(stdout);
#endif
#ifdef COUNT_FRAMES
    FramesDisplayed++;
#endif
  }

  void main(void)
  {
    long long Frame;

    while(1)
    {
      Frame = VideoIn->receive();
      display(Frame);
    }
  }

  SC_CTOR(Display)
  {
#ifdef COUNT_FRAMES
    FramesDisplayed = 0;
#endif
    SC_THREAD(main);
  }
};

SC_MODULE(SpeakerL)
{
  sc_port<i_long_long_receiverL> AudioIn;
#ifdef COUNT_FRAMES
  unsigned FramesPlayed;
#endif

  char Channel;

  void sound(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    putchar(Channel); // output the sound
    fflush(stdout);
#endif
#ifdef COUNT_FRAMES
    FramesPlayed++;
#endif
  }

  void main(void)
  {
    long long Frame;

    while(1)
    {
      Frame = AudioIn->receive();
      sound(Frame);
    }
  }

  SpeakerL(sc_module_name name, char chn): sc_module(name), Channel(chn)
  {
#ifdef COUNT_FRAMES
    FramesPlayed = 0;
#endif
    SC_THREAD(main);
  }
  SC_HAS_PROCESS(SpeakerL);
};


SC_MODULE(SpeakerR)
{
  sc_port<i_long_long_receiverR> AudioIn;
#ifdef COUNT_FRAMES
  unsigned FramesPlayed;
#endif

  char Channel;

  void sound(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    putchar(Channel); // output the sound
    fflush(stdout);
#endif
#ifdef COUNT_FRAMES
    FramesPlayed++;
#endif
  }

  void main(void)
  {
    long long Frame;

    while(1)
    {
      Frame = AudioIn->receive();
      sound(Frame);
    }
  }

  SpeakerR(sc_module_name name, char chn): sc_module(name), Channel(chn)
  {
#ifdef COUNT_FRAMES
    FramesPlayed = 0;
#endif
    SC_THREAD(main);
  }
  SC_HAS_PROCESS(SpeakerR);
};

SC_MODULE(Monitor)
{
  sc_port<i_long_long_receiverV> Video;
  sc_port<i_long_long_receiverL> AudioLeft;
  sc_port<i_long_long_receiverR> AudioRight;

  Display screen;
  SpeakerL speakerL;
  SpeakerR speakerR;

  void before_end_of_elaboration()
  {
    screen.VideoIn.bind(Video);
    speakerL.AudioIn.bind(AudioLeft);
    speakerR.AudioIn.bind(AudioRight);
  }

  SC_CTOR(Monitor):
        screen("screen"),
        speakerL("speakerL", 'L'),
        speakerR("speakerR", 'R')
  { }
};

SC_MODULE(Top)
{
  c_long_long_queueVS VideoStream;
  c_long_long_queueLS AudioStreamL;
  c_long_long_queueRS AudioStreamR;
  c_long_long_queueV Video;
  c_long_long_queueL AudioLeft;
  c_long_long_queueR AudioRight;

  Stimulus       stimulus;
  DUT            dut;
  Monitor        monitor;

  void before_end_of_elaboration()
  {
    stimulus.VideoStream.bind(VideoStream);
    stimulus.AudioStreamL.bind(AudioStreamL);
    stimulus.AudioStreamR.bind(AudioStreamR);

    dut.VideoIn.bind(VideoStream);
    dut.AudioInL.bind(AudioStreamL);
    dut.AudioInR.bind(AudioStreamR);
    dut.VideoOut.bind(Video);
    dut.AudioOutLeft.bind(AudioLeft);
    dut.AudioOutRight.bind(AudioRight);

    monitor.Video.bind(Video);
    monitor.AudioLeft.bind(AudioLeft);
    monitor.AudioRight.bind(AudioRight);
  }

  SC_CTOR(Top)
  : VideoStream("VideoStream", BUFSIZE)
  , AudioStreamL("AudioStreamL", BUFSIZE)
  , AudioStreamR("AudioStreamR", BUFSIZE)
  , Video("Video", BUFSIZE)
  , AudioLeft("AudioLeft", BUFSIZE)
  , AudioRight("AudioRight", BUFSIZE)
  , stimulus("stimulus")
  , dut("dut")
  , monitor("monitor")
  { }
};

int sc_main(int argc, char* argv[])
{
  Top top("top");
#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
#endif
  printf("Playing...\n");
  sc_start();
#ifdef PRINT_FRAMES
  printf("\n");
#endif
#ifdef COUNT_FRAMES
  printf("%u video frames displayed\n", top.monitor.screen.FramesDisplayed);
  printf("%u left audio frames played\n", top.monitor.speakerL.FramesPlayed);
  printf("%u right audio frames played\n", top.monitor.speakerR.FramesPlayed);
  assert(top.monitor.screen.FramesDisplayed ==
		sc_time(CLIP_LENGTH).value()/sc_time(VIDEO_DELAY).value()+1);
  assert(top.monitor.speakerL.FramesPlayed ==
		sc_time(CLIP_LENGTH).value()/sc_time(AUDIO_DELAY).value()+1);
  assert(top.monitor.speakerR.FramesPlayed ==
		sc_time(CLIP_LENGTH).value()/sc_time(AUDIO_DELAY).value()+1);
#endif
#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
  assert(sc_time_stamp() >= sc_time(CLIP_LENGTH));
#endif
  printf("Done!\n");
  return 0;
}

// EOF

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
