// synchronous_conflict.cc: regression test for parallel execution
// RD, 04/12/17

#include <stdio.h>
#include <assert.h>

#include "systemc.h"

#define ITERATIONS 10
#define WORKLOAD   10000000

int GlobalVar = 0;	// the conflict!

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

SC_MODULE(Worker1)
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(1, SC_NS);
	    Count++;
	    GlobalVar = 1;		// SystemC guarantees
	    w = Work(WORKLOAD);		// that this thread executes
	    assert(GlobalVar == 1);	// without interruption!
	}
	assert(Count == ITERATIONS);
    }

    SC_CTOR(Worker1)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker1);
	SC_THREAD(main);
    }
};

SC_MODULE(Worker2)
{
    unsigned Count;
    unsigned long w;

    void main(void)
    {
	for(int i=0; i<ITERATIONS; i++)
	{
	    wait(1, SC_NS);
	    Count++;
	    GlobalVar = 2;		// SystemC guarantees
	    w = Work(WORKLOAD);		// that this thread executes
	    assert(GlobalVar == 2);	// without interruption!
	}
	assert(Count == ITERATIONS);
    }

    SC_CTOR(Worker2)
    {
	Count = 0;
	SC_HAS_PROCESS(Worker2);
	SC_THREAD(main);
    }
};

SC_MODULE(TOP)
{
    Worker1 w1;
    Worker2 w2;

    SC_CTOR(TOP)
    : w1("w1"), w2("w2")
    {
    }
};

TOP top("top");

int sc_main(int argc, char* argv[])
{
    printf("synchronous_conflict starts\n");
    sc_start();
    assert(top.w1.Count == ITERATIONS);
    assert(top.w2.Count == ITERATIONS);
    assert(GlobalVar == 1 || GlobalVar == 2);
    printf("synchronous_conflict done.\n");
    printf("[potential speedup over SEQ: OOO 1x, NPD 1x, SYN 1x]\n");
    return 0;
}

// EOF

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
