#include <iostream>

#include <libgen.h>
#include <cstring>

#include "risc.h"

extern DynamicModule *ast_root;

int main(int argc, char* argv[])
{
  std::cout << std::endl << "start processing file ..." << std::endl;

  risc::Design d(argc,argv);

  if(ast_root) {
    ast_root->print_tree(0);
  } else {
    std::cout << "No root defined" << std::endl;
  }

  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  for(risc::HierarchicalChannelVectorIter
      iter_channel  = d.hierarchical_channel_.begin();
      iter_channel != d.hierarchical_channel_.end();
      iter_channel++) {

    std::cout << "Channel: " <<  (*iter_channel)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread  = (*iter_channel)->threads_.begin();
        iter_thread != (*iter_channel)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;
      sg.add_function(*iter_thread);
    }

    std::cout << std::endl;
  }

  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    std::cout << "Module: " <<  (*iter_module)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread  = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;
      sg.add_function(*iter_thread);
    }

    std::cout << std::endl;
  }

  std::cout << "+++++ interfaces +++++" << std::endl;
  std::list<SgClassDefinition*> interfaces
    = risc::tools::get_all_used_interfaces();

  for(std::list<SgClassDefinition*>::iterator
      iter  = interfaces.begin();
      iter != interfaces.end();
      iter++) {

    std::cout << SageInterface::get_name(*iter) << std::endl;

    std::vector<SgNode*> func_decls
      = NodeQuery::querySubTree(*iter, V_SgFunctionDeclaration);

    for(std::vector<SgNode*>::iterator
        func_decl_iter  = func_decls.begin();
        func_decl_iter != func_decls.end();
        func_decl_iter++) {

      SgFunctionDeclaration *func_decl
        = isSgFunctionDeclaration(*func_decl_iter);

      std::cout << "  " << SageInterface::get_name(*func_decl_iter)
                << ", Line: "
                << func_decl->get_file_info()->get_line()
                << std::endl;
    }
  }

  std::cout << "+++++++ annotated channel interfaces implementation +++++++" << std::endl;
  std::list<SgFunctionDeclaration*> implemented_interfaces
    = risc::tools::get_all_implemented_interface_functions(true, true);

  assert(implemented_interfaces.size() == 3);

  for(std::list<SgFunctionDeclaration*>::iterator
      iter  = implemented_interfaces.begin();
      iter != implemented_interfaces.end();
      iter++) {

    std::cout << SageInterface::get_name(*iter)
              << ", Line: " << (*iter)->get_file_info()->get_line()
              << std::endl;
  }

  std::cout << "+++++++ port calls +++++++" << std::endl;

  std::list<SgFunctionCallExp*> port_calls
    = risc::tools::get_all_port_calls_to_annotated_functions(sg);

  assert(port_calls.size() == 3);

  for(std::list<SgFunctionCallExp*>::iterator
      iter  = port_calls.begin();
      iter != port_calls.end();
      iter++) {

    std::cout << (*iter)->unparseToString()
              << ", Line: " << (*iter)->get_file_info()->get_line()
              << std::endl;
  }

  std::cout << "++++++++++++++++++++++++++++" << std::endl;

  std::cout << std::endl << "end processing file ..." << std::endl;

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
